import pathlib
import datetime
import os

import pytest
import piexif

import pixe


@pytest.fixture
def src_file(src_path):
    return src_path.joinpath("red.jpg")


def test_calc_checksum(src_file):
    expected_checksum = "1cdef99be68dbdea159ec6fa8469b41ca13e9e6f"

    calculated_checksum = pixe._calc_checksum(src_file)

    assert expected_checksum == calculated_checksum


def test_extract_date(src_file):
    expected_date = datetime.datetime(2020, 3, 21, 3, 13, 12)

    extracted_date = pixe._extract_date(src_file)

    assert expected_date == extracted_date


def test_new_tag_owner(src_file):
    path_str = str(src_file)
    orig_exif = piexif.load(path_str)

    new_exif = piexif.load(pixe._new_tags(src_file, owner="Joe User"))

    assert orig_exif != new_exif
    assert new_exif["Exif"][0xa430] == b"Joe User"


def test_new_tag_copyright(src_file):
    path_str = str(src_file)
    orig_exif = piexif.load(path_str)

    new_exif = piexif.load(
        pixe._new_tags(src_file, copyright="Copyright 2023 Joe User.")
    )

    assert orig_exif != new_exif
    assert new_exif["0th"][piexif.ImageIFD.Copyright] == b"Copyright 2023 Joe User."


def test_process_file(src_file, dst_path):
    expected_file = pathlib.Path(dst_path).joinpath(
        "2020", "3", "20200321_031312_1cdef99be68dbdea159ec6fa8469b41ca13e9e6f.jpg"
    )

    pixe._process_file(src_file, dst_path)

    assert expected_file.exists()


def test_process_file_no_date(src_path, dst_path):
    src_file = src_path.joinpath("chocolate.jpg")
    new_file = pathlib.Path(dst_path).joinpath(
        "1902", "2", "19020220_000000_2a00d2b48e39f63cf834d4f7c50b2c1aa3b43a9c.jpg"
    )

    pixe._process_file(src_file, dst_path)

    assert new_file.exists()



import pathlib
import os
import datetime

import pytest
from click.testing import CliRunner
import piexif

import pixe


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def src_file(src_path):
    return src_path.joinpath("red.jpg")


def test_single_file(runner, src_file, dst_path):
    dest_file = pathlib.Path(dst_path).joinpath(
        "2020", "3", "20200321_031312_1cdef99be68dbdea159ec6fa8469b41ca13e9e6f.jpg"
    )

    results = runner.invoke(pixe.cli, f"--dest {dst_path} {src_file}")

    assert results.exit_code == 0
    assert dest_file.exists()


def test_single_file_no_exist(runner, dst_path):
    results = runner.invoke(pixe.cli, f"--dest {dst_path} this/file/really/does/not/exist")

    assert results.exit_code == 2


def test_single_file_bad(runner, dst_path):
    results = runner.invoke(pixe.cli, f"--dest {dst_path} /dev/zero")

    assert results.exit_code == 2


def test_single_file_duplicate(runner, src_file, dst_path):
    import_time = datetime.datetime.now()
    dest_file = dst_path.joinpath(
        "dups",
        import_time.strftime("%Y%m%d_%H%M%S"),
        "2020",
        "3",
        "20200321_031312_1cdef99be68dbdea159ec6fa8469b41ca13e9e6f.jpg",
    )
    runner.invoke(pixe.cli, f"--dest {dst_path} {src_file}")
    results = runner.invoke(pixe.cli, f"--dest {dst_path} {src_file}")

    assert results.exit_code == 0
    assert dest_file.exists()


def test_single_file_move(runner, src_file, dst_path):
    dest_file = dst_path.joinpath("2020", "3", "20200321_031312_1cdef99be68dbdea159ec6fa8469b41ca13e9e6f.jpg")

    results = runner.invoke(pixe.cli, f"--move --dest {dst_path} {src_file}")

    assert results.exit_code == 0
    assert dest_file.exists()
    assert not src_file.exists()


def test_single_file_copy_tagged(runner, src_path, dst_path):
    src_file = src_path.joinpath("dark/darkturquoise.jpg")
    dest_file = dst_path.joinpath("2020", "12", "20201209_015501_a810b8552a4acf4e13164a74aab3016e583cc93e.jpg")
    old_checksum = dest_file.stem.split("_")[2]

    results = runner.invoke(
        pixe.cli, f"--copy --owner 'Joe User' --copyright 'Copyright 2020 Joe User.' --dest {dst_path} {src_file}"
    )
    src_exif = piexif.load(str(src_file))
    dst_exif = piexif.load(str(dest_file))
    new_checksum = pixe._calc_checksum(dest_file)

    assert results.exit_code == 0
    assert dest_file.exists()
    assert src_exif != dst_exif
    assert dst_exif["Exif"][0xA430] == b"Joe User"
    assert dst_exif["0th"][piexif.ImageIFD.Copyright] == b"Copyright 2020 Joe User."
    assert old_checksum == new_checksum


def test_files_parallel(runner, src_path, dst_path):
    results = runner.invoke(pixe.cli, f"--dest {dst_path} {src_path}")

    assert results.exit_code == 0
    assert dst_path.joinpath("2021", "12", "20211222_153825_d05cae67991384d221e95ae8b30994ce186695ed.jpg").exists()


def test_files_serial(runner, src_path, dst_path):
    results = runner.invoke(pixe.cli, f"--serial --dest {dst_path} {src_path}")

    assert results.exit_code == 0
    assert dst_path.joinpath("2021", "12", "20211222_153825_d05cae67991384d221e95ae8b30994ce186695ed.jpg").exists()


def test_files_recurse(runner, src_path, dst_path):
    results = runner.invoke(pixe.cli, f"--recurse --dest {dst_path} {src_path}")

    assert results.exit_code == 0
    assert dst_path.joinpath("2022", "2", "20220226_001821_476bf667385499407e1405f5909f88875dab1873.jpg").exists()
    assert dst_path.joinpath("2018", "3", "20180319_100139_b78473e5c10d8fd945dd1eee9da7a82320d464d1.jpg").exists()
    assert dst_path.joinpath("2021", "12", "20211222_153825_d05cae67991384d221e95ae8b30994ce186695ed.jpg").exists()



import hashlib
import io
import datetime
import pathlib
import multiprocessing
import shutil
import re
import os
import fnmatch
import pkg_resources

import click
import PIL.Image
import piexif

# Using a date that shouldn't appear in our collection, but that also isn't a common default.
# In this case, Ansel Adams birthday.
ERROR_DATE = datetime.datetime(1902, 2, 20)

# store a datetime of when this run began
START_TIME = datetime.datetime.now()

# Package introspection
PKG_RESOURCES = pkg_resources.get_distribution('pixe')

def _calc_checksum(image_path: pathlib.Path, block_size: int = 8192) -> str:
    """
    Create a sha1 checksum of just the image data (no meta/exif).

    :param image_path: a path to an image to process
    :param block_size: the block size to use when chunking up the image data
    :return: a calculated hex digest
    """
    hasher = hashlib.sha1()
    img_io = io.BytesIO()

    # open the image file and save the image data portion as an io.BytesIO object
    with PIL.Image.open(image_path) as im:
        im.save(img_io, im.format)

    # reset the cursor
    img_io.seek(0)

    # chunk_size at a time, update our hash until complete
    while chunk := img_io.read(block_size):
        hasher.update(chunk)

    return hasher.hexdigest()


def _extract_date(image_path: pathlib.Path) -> datetime.datetime:
    """
    Extract the file creation date from EXIF information.

    :param image_path: the path to a specific image file
    :return: a datetime object representing the creation date of the image
    """
    # TODO: use piexif: exif_dict['Exif'][piexif.ExifIFD.DateTimeOriginal]
    with PIL.Image.open(image_path, "r") as im:
        try:
            # attempt to extract the creation date from EXIF tag 36867
            exif = im._getexif()
            cdate = datetime.datetime.strptime(exif[36867], "%Y:%m:%d %H:%M:%S")

        # the requested tag doesn't exist, use the ERROR_DATE global to signify such
        except KeyError:
            cdate = ERROR_DATE

        return cdate


def _new_tags(image_path: pathlib.Path, **kwargs) -> bytes:
    """
    Insert tags into image exif
    :param image_path: path to a given image
    :param owner: the camera owner
    :param copyright: a copyright string
    :return: a bytes object containing the new exif data
    """
    path_str = str(image_path)
    exif_dict = piexif.load(path_str)

    if "owner" in kwargs and kwargs.get("owner") != '':
        exif_dict["Exif"][0xa430] = kwargs.get("owner").encode("ascii")

    if "copyright" in kwargs and kwargs.get("copyright") != '':
        exif_dict["0th"][piexif.ImageIFD.Copyright] = kwargs.get("copyright").encode("ascii")

    return piexif.dump(exif_dict)


def _process_file(file_path: pathlib.Path, dest_str: str, move: bool = False, **kwargs) -> str:
    """
    process a single file
    :param file_path: the path to the file to process
    :param dest_str: where should the processed files be moved/copied to
    :param move: should the file be moved or copied
    :param kwargs: additional options (likely exif tags)
    :return: a string representing the operation that has been performed
    """
    cdate = _extract_date(file_path)
    cdate_str = cdate.strftime("%Y%m%d_%H%M%S")
    hash_str = _calc_checksum(file_path)
    filename = file_path.with_stem(f"{cdate_str}_{hash_str}").with_suffix(file_path.suffix.lower())
    dest_path = pathlib.Path(dest_str).joinpath(str(cdate.year), str(cdate.month))

    # if a similarly named file exists at the destination it means we have a duplicate file
    # prepend 'dups' and the START_TIME of this move process to the destination filepath
    if dest_path.joinpath(filename.name).exists():
        dest_path = pathlib.Path(dest_str).joinpath(
            f"dups/{START_TIME.strftime('%Y%m%d_%H%M%S')}",
            str(cdate.year),
            str(cdate.month),
        )
    dest_path.mkdir(parents=True, exist_ok=True)
    dest_file = dest_path.joinpath(filename.name)

    if move:
        shutil.move(file_path, dest_file)
    else:
        shutil.copy(file_path, dest_file)

    # pass **kwargs to _new_tags so that known tags can be inserted
    # into the file at its destination so we don't muck up the src_file
    # if a copy operation has been requested.
    piexif.insert(_new_tags(dest_file, **kwargs), str(dest_file))

    # return a string showing what file has been moved, so it can be displayed
    return f"{file_path} -> {dest_path.joinpath(filename.name)}"


def parallel_process_files(file_list: list, dest: str, move: bool, **kwargs):
    """
    process a list of files in parallel
    :param file_list: a list of files to be processed
    :param dest: the destination for the process operation
    :param move: is this a move or copy operation
    :param kwargs: additional options (likely exif tags)
    """
    pool = multiprocessing.Pool()
    for file in file_list:
        pool.apply_async(
            _process_file,
            args=(file, dest, move),
            kwds=kwargs,
            callback=(lambda res: print(res, flush=True)),
        )
    pool.close()
    pool.join()


def serial_process_files(file_list: list, dest: str, move: bool, **kwargs):
    """
    process a list of files serially
    :param file_list: a list of files to be processed
    :param dest: the destination for the process operation
    :param move: is this a move or copy operation
    :param kwargs: additional options (likely exif tags)
    """
    for file in file_list:
        print(_process_file(file, dest, move, **kwargs))


@click.command()
@click.argument("src")
@click.version_option(PKG_RESOURCES.version, '-v', '--version')
@click.option("--dest", "-d", default=".", help="desired destination")
@click.option(
    "--recurse",
    "-r",
    is_flag=True,
    default=False,
    help="recurse into sub-directories (default: off)",
)
@click.option(
    "--parallel/--serial",
    default=True,
    help="process files in parallel (default: --parallel)",
)
@click.option(
    "--move/--copy",
    "--mv/--cp",
    is_flag=True,
    default=False,
    help="move files into DEST rather than copying (default: --copy)",
)
@click.option(
    "--owner",
    default='',
    help="add camera owner to exif tags",
)
@click.option(
    "--copyright",
    default='',
    help="add copyright string to exif tags"
)
def cli(src: str, dest: str, recurse: bool, parallel: bool, move: bool, **kwargs):
    file_path = pathlib.Path(src)
    # jpg pattern (case insensitive)
    file_re = re.compile(fnmatch.translate("*.jpg"), re.IGNORECASE)
    if file_path.exists():
        if file_path.is_dir():
            file_list = []
            for root, dirs, files in os.walk(file_path, topdown=True):
                for file in files:
                    if file_re.match(file):
                        file_list.append(pathlib.Path(root).joinpath(file))
                if not recurse:
                    break

            if parallel:
                parallel_process_files(file_list, dest, move, **kwargs)
            else:
                serial_process_files(file_list, dest, move, **kwargs)

        elif file_path.is_file():
            print(_process_file(file_path, dest, move, **kwargs))
        else:
            raise click.exceptions.BadParameter(src)
    else:
        raise click.exceptions.BadParameter(src)
